#ifndef _K_SKETCH_H
#define _K_SKETCH_H

#include "hash.h"
#include "util.h"

using namespace std;

class KSketch {

 private:
  HashTable   **hashTbls;
  int           numTbls;    // H
  double        invFactor;  // 1.0/(K-1)
  ValueType     sum;        // total sum
  ValueType    *medBuf;     // buffer for computing median

 public:
  // Constructor
  KSketch(int numHashTbls, int inxBits);

  // Destructor
  ~KSketch();

  // Accessor
  ValueType Sum() const { return sum; }
  
  // Member Functions
  void Reset()
    {
      sum = 0;
      for ( int i = 0; i < numTbls; i++ )
        hashTbls[i]->Reset();
    }
  
  void AddValue(KeyType key, ValueType val)
    {
      for( int i = 0; i < numTbls; i++ )
        hashTbls[i]->AddValue(key, val);
      sum += val;
    }
  
  void RemoveValue(KeyType key, ValueType val)
    {
      for( int i = 0; i < numTbls; i++ )
        hashTbls[i]->RemoveValue(key, val);
      sum -= val;
    }
  
  ValueType EstValue(KeyType key)
    {
      for ( int i = 0; i < numTbls; i++ )
        medBuf[i] = hashTbls[i]->GetValue(key);

      ValueType med = quickmedian(medBuf, numTbls);

      return med + (med - sum) * invFactor;
    }
};

#endif
