#ifndef _REV_HASH_TABLE_48_H
#define _REV_HASH_TABLE_48_H

#include <math.h>
#include <iostream>
#include <vector>
#include <list>
#include "globals.h"

using namespace std;

class RevHashTable48 {

  // private:
 public:
  vector<vector<unsigned char> > lookupTable; // IP key sections -> hash index
  vector<ValueType>              valueTable;  // actual data goes here
  vector<vector<unsigned char> > revLookupTable[NUM_DIVISIONS48];

  unsigned int numBins;
  unsigned int bitsPerDiv;
  unsigned int binsPerDiv;

  // Helper Functions
  void PopulateLookupTable(int div, int seed);
  
 public:
  RevHashTable48(int seed, int inxBits);  // Constructor

  void AddValue(KeyType48 key, ValueType val);
  void RemoveValue(KeyType48 key, ValueType val);
  ValueType GetValue(KeyType48 key);

  unsigned int GetNumBins(void) {return numBins;}

  IndexType GetIndex(KeyType48 key);

  void GetPossibleKeys(list<unsigned char> &keys, int div, IndexType bin);

  // Change detection compares the absolute value of the difference of each
  // of the bins to the threshold. It returns a sorted vector containing
  // the indices of the bins exceeding the threshold.
  void ChangeDetection(list<IndexType> &changes, const RevHashTable48 &otherTbl, 
		       double threshold);

  double FindThreshold(const RevHashTable48 &otherTbl,
		       int numAnoms, double minThreshold);
};

#endif
