/* -*-  Mode:C++; c-basic-offset:8; tab-width:8; indent-tabs-mode:t -*- */
// $Header: /home/zhichun/Research/CVS/RevSketch/util.cpp,v 1.1.1.1 2005/01/13 17:22:28 zhichun Exp $

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <ctype.h>
#include <sys/time.h>

#include "util.h"

// Generates a 32-bit random string
// Note that rand() is not random enough -- it has a period of 2^32.
uint32 rand32bit()
	{
	uint32 base = 0;
	uint32 i;

	for (i=1; i<=2; i++)
		base = (base<<16) | rand();
	return(base);
	}

// Generates a 64-bit random string
// Note that rand() is not random enough -- it has a period of 2^32.
uint64 rand64bit()
	{
	uint64 base = 0;
	uint32 i;

	for (i=1; i<=4; i++)
		base = (base<<16) | rand();
	return(base);
	}

void message(const char* msg)
	{
	message("%s", msg);
	}

void message(const char* fmt, const char* arg)
	{
	fprintf(stderr, fmt, arg);
	fprintf(stderr, "\n");
	}

void internal_error(const char* msg)
	{
	internal_error("%s", msg);
	}

void internal_error(const char* fmt, const char* arg)
	{
	fprintf(stderr, "internal error: ");
	fprintf(stderr, fmt, arg);
	fprintf(stderr, "\n");
	abort();
	exit(1);
	}

double current_time()
	{
	struct timeval tv;
	if ( gettimeofday(&tv, 0) < 0 )
		internal_error("gettimeofday failed in current_time()");

	return double(tv.tv_sec) + double(tv.tv_usec) / 1e6;
	}

/* ----------------- prototypes for static functions --------------- */
/* median finding routings */
val_t quickselect(val_t arr[], int n);
val_t opt_med3(val_t *p);
val_t opt_med5(val_t *p);
val_t opt_med7(val_t *p);
val_t opt_med9(val_t *p);
val_t opt_med25(val_t *p);

/* ------------------- function implementation ------------------------ */

/* wrapper function */
val_t quickmedian(val_t *a, int n)
{
  val_t median;
  switch(n) 
    {
    case 1:	median = a[0]; break;
    case 2:	median = (a[0]+a[1])/2; break;
    case 3:	median = opt_med3(a); break;
    case 5:	median = opt_med5(a); break;
    case 7:	median = opt_med7(a); break;
    case 9:	median = opt_med9(a); break;
    case 25:	median = opt_med25(a); break;
    default:	median = quickselect(a, n); break;
    }
  return median;
}

/* ------------------- utility function ------------------------ */

/*
 * This quickselect() routine is based on the algorithm described in
 * "Numerical recipes in C", Second Edition,
 * Cambridge University Press, 1992, Section 8.5, ISBN 0-521-43108-5
 * This code by Nicolas Devillard - 1998. Public domain.
 *
 * Further copyright information can be found at:
 *	http://www.nr.com/infotop.html#distinfo
 * It seems to be fine if we bundle the code inside executable, or don't
 *	release the software.
 */
#define ELEM_SWAP(a,b) { register val_t t=(a);(a)=(b);(b)=t; }
val_t quickselect(val_t arr[], int n)
{
  int low, high, median;
  int middle, ll, hh;
  low = 0 ; high = n-1 ; median = (low + high) / 2;
  for (;;) {
    if (high <= low) /* One element only */
      return arr[median] ;
    if (high == low + 1) { /* Two elements only */
      if (arr[low] > arr[high])
	ELEM_SWAP(arr[low], arr[high]) ;
      return arr[median] ;
    }

    /* Find median of low, middle and high items; swap into position low */
    middle = (low + high) / 2;
    if (arr[middle] > arr[high]) ELEM_SWAP(arr[middle], arr[high]) ;
    if (arr[low] > arr[high]) ELEM_SWAP(arr[low], arr[high]) ;
    if (arr[middle] > arr[low]) ELEM_SWAP(arr[middle], arr[low]) ;

    /* Swap low item (now in position middle) into position (low+1) */
    ELEM_SWAP(arr[middle], arr[low+1]) ;

    /* Nibble from each end towards middle, swapping items when stuck */
    ll = low + 1;
    hh = high;
    for (;;) {
      do ll++; while (arr[low] > arr[ll]) ;
      do hh--; while (arr[hh] > arr[low]) ;
      if (hh < ll)
	break;
      ELEM_SWAP(arr[ll], arr[hh]) ;
    }

    /* Swap middle item (in position low) back into correct position */
    ELEM_SWAP(arr[low], arr[hh]) ;
    
    /* Re-set active partition */
    if (hh <= median)
      low = ll;
    if (hh >= median)
      high = hh - 1;
  }
}
#undef ELEM_SWAP

/*
 * The following routines have been built from knowledge gathered
 * around the Web. I am not aware of any copyright problem with
 * them, so use it as you want.
 * N. Devillard - 1998
 * http://ndevilla.free.fr/median/median.pdf
 */
typedef val_t pixelvalue ;
#define PIX_SORT(a,b) { if ((a)>(b)) PIX_SWAP((a),(b)); }
#define PIX_SWAP(a,b) { register pixelvalue temp=(a);(a)=(b);(b)=temp; }
/*----------------------------------------------------------------------------
  Function : opt_med3()
  In : pointer to array of 3 pixel values
  Out : a pixelvalue
  Job : optimized search of the median of 3 pixel values
  Notice : found on sci.image.processing
  cannot go faster unless assumptions are made
  on the nature of the input signal.
  ---------------------------------------------------------------------------*/
pixelvalue opt_med3(pixelvalue * p)
{
  PIX_SORT(p[0],p[1]) ; PIX_SORT(p[1],p[2]) ; PIX_SORT(p[0],p[1]) ;
  return(p[1]) ;
}
/*----------------------------------------------------------------------------
  Function : opt_med5()
  In : pointer to array of 5 pixel values
  Out : a pixelvalue
  Job : optimized search of the median of 5 pixel values
  Notice : found on sci.image.processing
  cannot go faster unless assumptions are made
  on the nature of the input signal.
  ---------------------------------------------------------------------------*/
pixelvalue opt_med5(pixelvalue * p)
{
  PIX_SORT(p[0],p[1]) ; PIX_SORT(p[3],p[4]) ; PIX_SORT(p[0],p[3]) ;
  PIX_SORT(p[1],p[4]) ; PIX_SORT(p[1],p[2]) ; PIX_SORT(p[2],p[3]) ;
  PIX_SORT(p[1],p[2]) ; return(p[2]) ;
}
/*----------------------------------------------------------------------------
  Function : opt_med7()
  In : pointer to array of 7 pixel values
  Out : a pixelvalue
  Job : optimized search of the median of 7 pixel values
  Notice : found on sci.image.processing
  cannot go faster unless assumptions are made
  on the nature of the input signal.
  ---------------------------------------------------------------------------*/
pixelvalue opt_med7(pixelvalue * p)
{
  PIX_SORT(p[0], p[5]) ; PIX_SORT(p[0], p[3]) ; PIX_SORT(p[1], p[6]) ;
  PIX_SORT(p[2], p[4]) ; PIX_SORT(p[0], p[1]) ; PIX_SORT(p[3], p[5]) ;
  PIX_SORT(p[2], p[6]) ; PIX_SORT(p[2], p[3]) ; PIX_SORT(p[3], p[6]) ;
  PIX_SORT(p[4], p[5]) ; PIX_SORT(p[1], p[4]) ; PIX_SORT(p[1], p[3]) ;
  PIX_SORT(p[3], p[4]) ; return (p[3]) ;
}
/*----------------------------------------------------------------------------
  Function : opt_med9()
  In : pointer to an array of 9 pixelvalues
  Out : a pixelvalue
  Job : optimized search of the median of 9 pixelvalues
  Notice : in theory, cannot go faster without assumptions on the signal.
  Formula from:
  XILINX XCELL magazine, vol. 23 by John L. Smith
  The input array is modified in the process
  The result array is guaranteed to contain the median value
  in middle position, but other elements are NOT sorted.
  ---------------------------------------------------------------------------*/
pixelvalue opt_med9(pixelvalue * p)
{
  PIX_SORT(p[1], p[2]) ; PIX_SORT(p[4], p[5]) ; PIX_SORT(p[7], p[8]) ;
  PIX_SORT(p[0], p[1]) ; PIX_SORT(p[3], p[4]) ; PIX_SORT(p[6], p[7]) ;
  PIX_SORT(p[1], p[2]) ; PIX_SORT(p[4], p[5]) ; PIX_SORT(p[7], p[8]) ;
  PIX_SORT(p[0], p[3]) ; PIX_SORT(p[5], p[8]) ; PIX_SORT(p[4], p[7]) ;
  PIX_SORT(p[3], p[6]) ; PIX_SORT(p[1], p[4]) ; PIX_SORT(p[2], p[5]) ;
  PIX_SORT(p[4], p[7]) ; PIX_SORT(p[4], p[2]) ; PIX_SORT(p[6], p[4]) ;
  PIX_SORT(p[4], p[2]) ; return(p[4]) ;
}
/*----------------------------------------------------------------------------
  Function : opt_med25()
  In : pointer to an array of 25 pixelvalues
  Out : a pixelvalue
  Job : optimized search of the median of 25 pixelvalues
  Notice : in theory, cannot go faster without assumptions on the
  signal.
  Code taken from Graphic Gems.
  ---------------------------------------------------------------------------*/
pixelvalue opt_med25(pixelvalue * p)
{
  PIX_SORT(p[0], p[1]) ; PIX_SORT(p[3], p[4]) ; PIX_SORT(p[2], p[4]) ;
  PIX_SORT(p[2], p[3]) ; PIX_SORT(p[6], p[7]) ; PIX_SORT(p[5], p[7]) ;
  PIX_SORT(p[5], p[6]) ; PIX_SORT(p[9], p[10]) ; PIX_SORT(p[8], p[10]) ;
  PIX_SORT(p[8], p[9]) ; PIX_SORT(p[12], p[13]) ; PIX_SORT(p[11], p[13]) ;
  PIX_SORT(p[11], p[12]) ; PIX_SORT(p[15], p[16]) ; PIX_SORT(p[14], p[16]) ;
  PIX_SORT(p[14], p[15]) ; PIX_SORT(p[18], p[19]) ; PIX_SORT(p[17], p[19]) ;
  PIX_SORT(p[17], p[18]) ; PIX_SORT(p[21], p[22]) ; PIX_SORT(p[20], p[22]) ;
  PIX_SORT(p[20], p[21]) ; PIX_SORT(p[23], p[24]) ; PIX_SORT(p[2], p[5]) ;
  PIX_SORT(p[3], p[6]) ; PIX_SORT(p[0], p[6]) ; PIX_SORT(p[0], p[3]) ;
  PIX_SORT(p[4], p[7]) ; PIX_SORT(p[1], p[7]) ; PIX_SORT(p[1], p[4]) ;
  PIX_SORT(p[11], p[14]) ; PIX_SORT(p[8], p[14]) ; PIX_SORT(p[8], p[11]) ;
  PIX_SORT(p[12], p[15]) ; PIX_SORT(p[9], p[15]) ; PIX_SORT(p[9], p[12]) ;
  PIX_SORT(p[13], p[16]) ; PIX_SORT(p[10], p[16]) ; PIX_SORT(p[10], p[13]) ;
  PIX_SORT(p[20], p[23]) ; PIX_SORT(p[17], p[23]) ; PIX_SORT(p[17], p[20]) ;
  PIX_SORT(p[21], p[24]) ; PIX_SORT(p[18], p[24]) ; PIX_SORT(p[18], p[21]) ;
  PIX_SORT(p[19], p[22]) ; PIX_SORT(p[8], p[17]) ; PIX_SORT(p[9], p[18]) ;
  PIX_SORT(p[0], p[18]) ; PIX_SORT(p[0], p[9]) ; PIX_SORT(p[10], p[19]) ;
  PIX_SORT(p[1], p[19]) ; PIX_SORT(p[1], p[10]) ; PIX_SORT(p[11], p[20]) ;
  PIX_SORT(p[2], p[20]) ; PIX_SORT(p[2], p[11]) ; PIX_SORT(p[12], p[21]) ;
  PIX_SORT(p[3], p[21]) ; PIX_SORT(p[3], p[12]) ; PIX_SORT(p[13], p[22]) ;
  PIX_SORT(p[4], p[22]) ; PIX_SORT(p[4], p[13]) ; PIX_SORT(p[14], p[23]) ;
  PIX_SORT(p[5], p[23]) ; PIX_SORT(p[5], p[14]) ; PIX_SORT(p[15], p[24]) ;
  PIX_SORT(p[6], p[24]) ; PIX_SORT(p[6], p[15]) ; PIX_SORT(p[7], p[16]) ;
  PIX_SORT(p[7], p[19]) ; PIX_SORT(p[13], p[21]) ; PIX_SORT(p[15], p[23]) ;
  PIX_SORT(p[7], p[13]) ; PIX_SORT(p[7], p[15]) ; PIX_SORT(p[1], p[9]) ;
  PIX_SORT(p[3], p[11]) ; PIX_SORT(p[5], p[17]) ; PIX_SORT(p[11], p[17]) ;
  PIX_SORT(p[9], p[17]) ; PIX_SORT(p[4], p[10]) ; PIX_SORT(p[6], p[12]) ;
  PIX_SORT(p[7], p[14]) ; PIX_SORT(p[4], p[6]) ; PIX_SORT(p[4], p[7]) ;
  PIX_SORT(p[12], p[14]) ; PIX_SORT(p[10], p[14]) ; PIX_SORT(p[6], p[7]) ;
  PIX_SORT(p[10], p[12]) ; PIX_SORT(p[6], p[10]) ; PIX_SORT(p[6], p[17]) ;
  PIX_SORT(p[12], p[17]) ; PIX_SORT(p[7], p[17]) ; PIX_SORT(p[7], p[10]) ;
  PIX_SORT(p[12], p[18]) ; PIX_SORT(p[7], p[12]) ; PIX_SORT(p[10], p[18]) ;
  PIX_SORT(p[12], p[20]) ; PIX_SORT(p[10], p[20]) ; PIX_SORT(p[10], p[12]) ;
  return (p[12]);
}
#undef PIX_SORT
#undef PIX_SWAP
